"""Top-level package for fhir-kindling."""

__author__ = """Michael Graf"""
__email__ = 'michael.graf@uni-tuebingen.de'
__version__ = '0.7.2'

from .fhir_server import FhirServer
from .fhir_query import FHIRQuery
